// SDL2.0 Benchmark (based on testsprite2.c of SDL).
// Roberto Prieto - contact@megastormsystems.com
// MegaStorm Systems (c) 2013 - http://www.megastormsystems.com

#include "SDL/SDL.h"
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#define NUM_SPRITES    1000
#define NUM_SPRITES2    100
#define MAX_SPEED     1
#define STEP_TIME     10

static int num_sprites;
static SDL_Texture **sprites;
static SDL_Rect *positions;
static SDL_Rect *velocities;
static int sprite_w, sprite_h;
static SDL_BlendMode blendMode = SDL_BLENDMODE_BLEND;

typedef struct
{
    Uint32 flags;

    // Video info
    const char *videodriver;
    int display;
    const char *window_title;
    const char *window_icon;
    Uint32 window_flags;
    int window_x;
    int window_y;
    int window_w;
    int window_h;
    int window_minW;
    int window_minH;
    int window_maxW;
    int window_maxH;
    int depth;
    int refresh_rate;
    int num_windows;
    SDL_Window **windows;

    // Renderer info
    const char *renderdriver;
    Uint32 render_flags;
    SDL_bool skip_renderer;
    SDL_Renderer **renderers;

    // GL settings
    int gl_red_size;
    int gl_green_size;
    int gl_blue_size;
    int gl_alpha_size;
    int gl_buffer_size;
    int gl_depth_size;
    int gl_stencil_size;
    int gl_double_buffer;
    int gl_accum_red_size;
    int gl_accum_green_size;
    int gl_accum_blue_size;
    int gl_accum_alpha_size;
    int gl_stereo;
    int gl_multisamplebuffers;
    int gl_multisamplesamples;
    int gl_retained_backing;
    int gl_accelerated;
    int gl_major_version;
    int gl_minor_version;
} Config;

static void PrintRendererFlag(Uint32 flag)
{
    switch (flag) {
    case SDL_RENDERER_PRESENTVSYNC:
        printf("PresentVSync");
        break;
    case SDL_RENDERER_ACCELERATED:
        printf("Accelerated");
        break;
    default:
        printf("0x%8.8x", flag);
        break;
    }
}

static void PrintRenderer(SDL_RendererInfo * info)
{
    int i, count;

    printf("  Renderer %s:\n", info->name);

    printf("    Flags: 0x%8.8X", info->flags);
    printf(" (");
    count = 0;
    for(i = 0; i < sizeof(info->flags) * 8; ++i) 
    {
        Uint32 flag = (1 << i);
        if(info->flags & flag) 
        {
            if(count > 0) 
            {
                fprintf(stderr, " | ");
            }
            PrintRendererFlag(flag);
            ++count;
        }
    }
    printf(")\n");

    if(info->max_texture_width || info->max_texture_height) 
        printf("    Max Texture Size: %dx%d\n", info->max_texture_width, info->max_texture_height);
}


static void quit(Config *state, int rc)
{
    if (sprites) 
    {
        SDL_free(sprites);
    }
    if (positions) 
    {
        SDL_free(positions);
    }
    if (velocities) 
    {
        SDL_free(velocities);
    }
    // Quit state
    if(state->windows) SDL_free(state->windows);
    if(state->renderers) 
    {
        for(int i = 0; i < state->num_windows; ++i) 
        {
            if(state->renderers[i]) SDL_DestroyRenderer(state->renderers[i]);
        }
        SDL_free(state->renderers);
    }
    if(state->flags & SDL_INIT_VIDEO) SDL_VideoQuit();
      
    if(rc > 0) 
    {
        SDL_free(state);
        exit(rc);
    }
}
static SDL_Surface *LoadIcon(const char *file)
{
    SDL_Surface *icon;

    // Load the icon surface
    icon = SDL_LoadBMP(file);
    if(icon == NULL) 
    {
        printf("Couldn't load %s: %s\n", file, SDL_GetError());
        return (NULL);
    }

    if(icon->format->palette) 
    {
        SDL_SetColorKey(icon, 1, *((Uint8 *) icon->pixels));
    }
    return (icon);
}

SDL_bool Init(Config *state)
{
    int i, j, m, n, w, h;
    SDL_DisplayMode fullscreen_mode;

    if(state->flags & SDL_INIT_VIDEO) 
    {
        n = SDL_GetNumVideoDrivers();
        if(n == 0) printf("No built-in video drivers\n");
        else 
        {
            printf("Built-in video drivers:");
            for(i = 0; i < n; ++i) 
            {
                if(i > 0) 
                {
                    printf(",");
                }
                printf(" %s", SDL_GetVideoDriver(i));
            }
            printf("\n");
        }
       
        if(SDL_VideoInit(state->videodriver) < 0) 
        {
            printf("Couldn't initialize video driver: %s\n",SDL_GetError());
            return SDL_FALSE;
        }
        printf("Video driver: %s\n", SDL_GetCurrentVideoDriver());
        
        // Upload GL settings
        SDL_GL_SetAttribute(SDL_GL_RED_SIZE, state->gl_red_size);
        SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, state->gl_green_size);
        SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, state->gl_blue_size);
        SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, state->gl_alpha_size);
        SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, state->gl_double_buffer);
        SDL_GL_SetAttribute(SDL_GL_BUFFER_SIZE, state->gl_buffer_size);
        SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, state->gl_depth_size);
        SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, state->gl_stencil_size);
        SDL_GL_SetAttribute(SDL_GL_ACCUM_RED_SIZE, state->gl_accum_red_size);
        SDL_GL_SetAttribute(SDL_GL_ACCUM_GREEN_SIZE, state->gl_accum_green_size);
        SDL_GL_SetAttribute(SDL_GL_ACCUM_BLUE_SIZE, state->gl_accum_blue_size);
        SDL_GL_SetAttribute(SDL_GL_ACCUM_ALPHA_SIZE, state->gl_accum_alpha_size);
        SDL_GL_SetAttribute(SDL_GL_STEREO, state->gl_stereo);
        SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, state->gl_multisamplebuffers);
        SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, state->gl_multisamplesamples);
        if(state->gl_accelerated >= 0) SDL_GL_SetAttribute(SDL_GL_ACCELERATED_VISUAL,state->gl_accelerated);
        SDL_GL_SetAttribute(SDL_GL_RETAINED_BACKING, state->gl_retained_backing);
        if(state->gl_major_version) 
        {
            SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, state->gl_major_version);
            SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, state->gl_minor_version);
        }

        // Displays info
        SDL_Rect bounds;
        SDL_DisplayMode mode;
        int bpp;
        Uint32 Rmask, Gmask, Bmask, Amask;
        n = SDL_GetNumVideoDisplays();
        printf("Number of displays: %d\n", n);
        for(i = 0; i < n; ++i) 
        {
            printf("Display %d: %s\n", i, SDL_GetDisplayName(i));
            SDL_zero(bounds);
            SDL_GetDisplayBounds(i, &bounds);
            printf("Bounds: %dx%d at %d,%d\n", bounds.w, bounds.h, bounds.x, bounds.y);
            SDL_GetDesktopDisplayMode(i, &mode);
            SDL_PixelFormatEnumToMasks(mode.format, &bpp, &Rmask, &Gmask,&Bmask, &Amask);
            printf("  Current mode: %dx%d@%dHz, %d bits-per-pixel (%s)\n",mode.w, mode.h, mode.refresh_rate, bpp,SDL_GetPixelFormatName(mode.format));
            if(Rmask || Gmask || Bmask) 
            {
                printf("      Red Mask   = 0x%.8x\n", Rmask);
                printf("      Green Mask = 0x%.8x\n", Gmask);
                printf("      Blue Mask  = 0x%.8x\n", Bmask);
                if(Amask) printf("      Alpha Mask = 0x%.8x\n", Amask);
            }

            // Print available fullscreen video modes
            m = SDL_GetNumDisplayModes(i);
            if(m == 0) printf("No available fullscreen video modes\n");
            else 
            {
                /*printf("  Fullscreen video modes:\n");
                for(j = 0; j < m; ++j) 
                {
                    SDL_GetDisplayMode(i, j, &mode);
                    SDL_PixelFormatEnumToMasks(mode.format, &bpp, &Rmask,&Gmask, &Bmask, &Amask);
                    printf("    Mode %d: %dx%d@%dHz, %d bits-per-pixel (%s)\n",j, mode.w, mode.h, mode.refresh_rate, bpp,SDL_GetPixelFormatName(mode.format));
                    if(Rmask || Gmask || Bmask)
                    {
                        printf("        Red Mask   = 0x%.8x\n", Rmask);
                        printf("        Green Mask = 0x%.8x\n", Gmask);
                        printf("        Blue Mask  = 0x%.8x\n", Bmask);
                        if(Amask) printf("        Alpha Mask = 0x%.8x\n", Amask);
                    }
                }*/
            }
        }

        // Renderer info
        SDL_RendererInfo info;
        n = SDL_GetNumRenderDrivers();
        if(n == 0) printf("No built-in render drivers\n");
        else 
        {
            printf("Built-in render drivers:\n");
            for(i = 0; i < n; ++i) 
            {
                SDL_GetRenderDriverInfo(i, &info);
                PrintRenderer(&info);
            }
        }

        // Set window
        SDL_zero(fullscreen_mode);
        switch (state->depth) 
        {
        case 8:
            fullscreen_mode.format = SDL_PIXELFORMAT_INDEX8;
            break;
        case 15:
            fullscreen_mode.format = SDL_PIXELFORMAT_RGB555;
            break;
        case 16:
            fullscreen_mode.format = SDL_PIXELFORMAT_RGB565;
            break;
        case 24:
            fullscreen_mode.format = SDL_PIXELFORMAT_RGB24;
            break;
        default:
            fullscreen_mode.format = SDL_PIXELFORMAT_RGB888;
            break;
        }
        fullscreen_mode.refresh_rate = state->refresh_rate;

        state->windows = (SDL_Window **) SDL_malloc(state->num_windows * sizeof(*state->windows));
        state->renderers = (SDL_Renderer **) SDL_malloc(state->num_windows * sizeof(*state->renderers));
        if (!state->windows || !state->renderers) 
        {
            printf("Out of memory!\n");
            return SDL_FALSE;
        }
        for(i = 0; i < state->num_windows; ++i) 
        {
            char title[1024];

            if (state->num_windows > 1) 
            {
                SDL_snprintf(title, SDL_arraysize(title), "%s %d", state->window_title, i + 1);
            } 
            else 
            {
                SDL_strlcpy(title, state->window_title, SDL_arraysize(title));
            }
            state->windows[i] = SDL_CreateWindow(title, state->window_x, state->window_y,state->window_w, state->window_h, state->window_flags);
            if(!state->windows[i]) 
            {
                printf("Couldn't create window: %s\n", SDL_GetError());
                return SDL_FALSE;
            }
            if(state->window_minW || state->window_minH) 
            {
                SDL_SetWindowMinimumSize(state->windows[i], state->window_minW, state->window_minH);
            }
            if(state->window_maxW || state->window_maxH) 
            {
                SDL_SetWindowMaximumSize(state->windows[i], state->window_maxW, state->window_maxH);
            }
            SDL_GetWindowSize(state->windows[i], &w, &h);
            if (!(state->window_flags & SDL_WINDOW_RESIZABLE) && (w != state->window_w || h != state->window_h)) 
            {
                printf("Window requested size %dx%d, got %dx%d\n", state->window_w, state->window_h, w, h);
                state->window_w = w;
                state->window_h = h;
            }
            if(SDL_SetWindowDisplayMode(state->windows[i], &fullscreen_mode) < 0) 
            {
                printf("Can't set up fullscreen display mode: %s\n", SDL_GetError());
                return SDL_FALSE;
            }

            if(state->window_icon) 
            {
                SDL_Surface *icon = LoadIcon(state->window_icon);
                if(icon) 
                {
                    SDL_SetWindowIcon(state->windows[i], icon);
                    SDL_FreeSurface(icon);
                }
            }

            SDL_ShowWindow(state->windows[i]);

            state->renderers[i] = NULL;

            if(!state->skip_renderer && (state->renderdriver || !(state->window_flags & SDL_WINDOW_OPENGL))) 
            {
                m = -1;
                if(state->renderdriver) 
                {
                    SDL_RendererInfo info;
                    n = SDL_GetNumRenderDrivers();
                    for(j = 0; j < n; ++j) 
                    {
                        SDL_GetRenderDriverInfo(j, &info);
                        if (SDL_strcasecmp(info.name, state->renderdriver) == 0) 
                        {
                            m = j;
                            break;
                        }
                    }
                    if(m == n) 
                    {
                        printf("Couldn't find render driver named %s", state->renderdriver);
                        return SDL_FALSE;
                    }
                }
                state->renderers[i] = SDL_CreateRenderer(state->windows[i],m, state->render_flags);
                if(!state->renderers[i]) 
                {
                    printf("Couldn't create renderer: %s\n", SDL_GetError());
                    return SDL_FALSE;
                }
                SDL_RendererInfo info;
                printf("Current renderer:\n");
                SDL_GetRendererInfo(state->renderers[i], &info);
                PrintRenderer(&info);
            }
        }
    }

    return SDL_TRUE;
}


int LoadSprite(Config *state, char *file)
{
    int i;
    SDL_Surface *temp;

    /* Load the sprite image */
    temp = SDL_LoadBMP(file);
    if (temp == NULL) {
        fprintf(stderr, "Couldn't load %s: %s", file, SDL_GetError());
        return (-1);
    }
    sprite_w = temp->w;
    sprite_h = temp->h;

    /* Set transparent pixel as the pixel at (0,0) */
    if (temp->format->palette) {
        SDL_SetColorKey(temp, 1, *(Uint8 *) temp->pixels);
    } else {
        switch (temp->format->BitsPerPixel) {
        case 15:
            SDL_SetColorKey(temp, 1, (*(Uint16 *) temp->pixels) & 0x00007FFF);
            break;
        case 16:
            SDL_SetColorKey(temp, 1, *(Uint16 *) temp->pixels);
            break;
        case 24:
            SDL_SetColorKey(temp, 1, (*(Uint32 *) temp->pixels) & 0x00FFFFFF);
            break;
        case 32:
            SDL_SetColorKey(temp, 1, *(Uint32 *) temp->pixels);
            break;
        }
    }

    /* Create textures from the image */
    for (i = 0; i < state->num_windows; ++i) 
    {
        SDL_Renderer *renderer = state->renderers[i];
        sprites[i] = SDL_CreateTextureFromSurface(renderer, temp);
        if (!sprites[i]) {
            fprintf(stderr, "Couldn't create texture: %s\n", SDL_GetError());
            SDL_FreeSurface(temp);
            return (-1);
        }
        SDL_SetTextureBlendMode(sprites[i], blendMode);
    }
    SDL_FreeSurface(temp);

    /* We're ready to roll. :) */
    return (0);
}

void MoveSprites(SDL_Renderer * renderer, SDL_Texture * sprite)
{
    int i, n;
    SDL_Rect viewport;
    SDL_Rect *position, *velocity;

    // Query the sizes
    SDL_RenderGetViewport(renderer, &viewport);
   
    // Grey background
    SDL_SetRenderDrawColor(renderer, 0xA0, 0xA0, 0xA0, 0xFF);
    SDL_RenderClear(renderer);

    // Move the sprite, bounce at the wall, and draw
    n = 0;
    for (i = 0; i < num_sprites; ++i) {
        position = &positions[i];
        velocity = &velocities[i];
        position->x += velocity->x;
        if ((position->x < 0) || (position->x >= (viewport.w - sprite_w))) {
            velocity->x = -velocity->x;
            position->x += velocity->x;
        }
        position->y += velocity->y;
        if ((position->y < 0) || (position->y >= (viewport.h - sprite_h))) {
            velocity->y = -velocity->y;
            position->y += velocity->y;
        }

        // Blit the sprite onto the screen
        SDL_RenderCopy(renderer, sprite, NULL, position);
    }

    // Update the screen!
    SDL_RenderPresent(renderer);
}


int main(int argc, char *argv[])
{
    int i, done;
    Uint32 then, now, frames;

    // Output file
    FILE *fp = fopen("sdl20.output","wb");
    double fps,fps2;
    SDL_RendererInfo ren;
    int iRender = 0;

    // Create the state
    Config *state = (Config *)SDL_calloc(1, sizeof(*state));
    if(!state) 
    {
        SDL_OutOfMemory();
        return 3;
    }
    state->flags = SDL_INIT_VIDEO;
    state->window_title = "SDL 2.0 Benchmarks";
    state->window_flags = SDL_WINDOW_FULLSCREEN;
    state->window_x = SDL_WINDOWPOS_UNDEFINED;
    state->window_y = SDL_WINDOWPOS_UNDEFINED;
    state->window_w = 800;
    state->window_h = 600;
    state->num_windows = 1;  
    // Set some very sane GL defaults
    state->gl_red_size = 3;
    state->gl_green_size = 3;
    state->gl_blue_size = 2;
    state->gl_alpha_size = 0;
    state->gl_buffer_size = 0;
    state->gl_depth_size = 16;
    state->gl_stencil_size = 0;
    state->gl_double_buffer = 0;
    state->gl_accum_red_size = 0;
    state->gl_accum_green_size = 0;
    state->gl_accum_blue_size = 0;
    state->gl_accum_alpha_size = 0;
    state->gl_stereo = 0;
    state->gl_multisamplebuffers = 0;
    state->gl_multisamplesamples = 0;
    state->gl_retained_backing = 1;
    state->gl_accelerated = -1;
	   
REINIT_SDL:
    // Initialize parameters
    num_sprites = NUM_SPRITES;

    // Initialize 
    if(!Init(state)) quit(state,2);
    
    // Create the windows, initialize the renderers, and load the textures
    sprites = (SDL_Texture **) SDL_malloc(state->num_windows * sizeof(*sprites));
    if(!sprites)
    {
        fprintf(stderr, "Out of memory!\n");
        quit(state,2);
    }
    for(i = 0; i < state->num_windows; ++i) 
    {
        SDL_Renderer *renderer = state->renderers[i];
        SDL_SetRenderDrawColor(renderer, 0xA0, 0xA0, 0xA0, 0xFF);
        SDL_RenderClear(renderer);
    }
    if(LoadSprite(state,"icon.bmp") < 0) 
    {
        quit(state,2);
    }

    /* Allocate memory for the sprite info */
    positions = (SDL_Rect *) SDL_malloc(num_sprites * sizeof(SDL_Rect));
    velocities = (SDL_Rect *) SDL_malloc(num_sprites * sizeof(SDL_Rect));
    if(!positions || !velocities) 
    {
        fprintf(stderr, "Out of memory!\n");
        quit(state,2);
    }
    srand((unsigned int)time(NULL));
    for(i = 0; i < num_sprites; ++i) 
    {
        positions[i].x = rand() % (state->window_w - sprite_w);
        positions[i].y = rand() % (state->window_h - sprite_h);
        positions[i].w = sprite_w;
        positions[i].h = sprite_h;
        velocities[i].x = 0;
        velocities[i].y = 0;
        while (!velocities[i].x && !velocities[i].y) 
        {
            velocities[i].x = (rand() % (MAX_SPEED * 2 + 1)) - MAX_SPEED;
            velocities[i].y = (rand() % (MAX_SPEED * 2 + 1)) - MAX_SPEED;
        }
    }

    // ---Main render loop for NUM_SPRITES2---
    num_sprites = NUM_SPRITES2;
    frames = 0;	
	then = SDL_GetTicks();	
	done = 0;	
	while ( !done ) 
	{				
		++frames;
        for(i = 0; i < state->num_windows; ++i) 
        {
            MoveSprites(state->renderers[i], sprites[i]);
        }
		Uint32 iTime = SDL_GetTicks();
		if(((iTime-then)/1000)>STEP_TIME) { done=1;  }
    }
    // Print out some timing information
    now = SDL_GetTicks();    
    if(now > then) 
    {
        fps = ((double) frames * 1000) / (now - then);
        printf("Moving %d sprites at %2.2f frames per second\n", num_sprites,fps);
    }	
    // ------

    // ---Main render loop for NUM_SPRITES---
    num_sprites = NUM_SPRITES;
    frames = 0;	
	then = SDL_GetTicks();	
	done = 0;	
	while ( !done ) 
	{				
		++frames;
        for(i = 0; i < state->num_windows; ++i) 
        {
            MoveSprites(state->renderers[i], sprites[i]);
        }
		Uint32 iTime = SDL_GetTicks();
		if(((iTime-then)/1000)>STEP_TIME) { done=1;  }
    }
    // Print out some timing information
    now = SDL_GetTicks();    
    if(now > then) 
    {
        fps2 = ((double) frames * 1000) / (now - then);
        printf("Moving %d sprites at %2.2f frames per second\n\n", num_sprites,fps2);
    }
    // ------


    // Major loop through the 3 renderers
    switch(iRender)
    {
    case 0: 
        // Dump info of d3d
        SDL_GetRendererInfo(state->renderers[0],&ren);
        fprintf(fp,"<tr><td width=196 align=center height=19><b><font size=2 color=#000080>SDL 2.0 using %s</font></b></td>",ren.name);
	    fprintf(fp,"<td width=99 align=center height=19><font size=2 color=#800000>%2.2f</font></td>",fps);
	    fprintf(fp,"<td width=94 align=center height=19><font size=2 color=#800000>%2.2f</font></td>",fps2);
	    fprintf(fp,"<td width=105 align=center height=19><font size=2 color=#800000>--</font></td>");
	    fprintf(fp,"<td width=95 align=center height=19><font size=2 color=#800000>--</font></td></tr>");
        iRender = 1;
        quit(state,0);
        state->renderdriver = "opengl";
        goto REINIT_SDL;
        break;
    case 1: 
        // Dump info of opengl
        SDL_GetRendererInfo(state->renderers[0],&ren);
        fprintf(fp,"<tr><td width=196 align=center height=19><b><font size=2 color=#000080>SDL 2.0 using %s</font></b></td>",ren.name);
	    fprintf(fp,"<td width=99 align=center height=19><font size=2 color=#800000>%2.2f</font></td>",fps);
	    fprintf(fp,"<td width=94 align=center height=19><font size=2 color=#800000>%2.2f</font></td>",fps2);
	    fprintf(fp,"<td width=105 align=center height=19><font size=2 color=#800000>--</font></td>");
	    fprintf(fp,"<td width=95 align=center height=19><font size=2 color=#800000>--</font></td></tr>");
        iRender = 2;
        quit(state,0);
        state->renderdriver = "software";
        goto REINIT_SDL;
        break;
    case 2:
        // Dump info of software
        SDL_GetRendererInfo(state->renderers[0],&ren);
        fprintf(fp,"<tr><td width=196 align=center height=19><b><font size=2 color=#000080>SDL 2.0 using %s</font></b></td>",ren.name);
	    fprintf(fp,"<td width=99 align=center height=19><font size=2 color=#800000>%2.2f</font></td>",fps);
	    fprintf(fp,"<td width=94 align=center height=19><font size=2 color=#800000>%2.2f</font></td>",fps2);
	    fprintf(fp,"<td width=105 align=center height=19><font size=2 color=#800000>--</font></td>");
	    fprintf(fp,"<td width=95 align=center height=19><font size=2 color=#800000>--</font></td></tr>");
    }            

    // End!    
    fclose(fp);
    quit(state,0);
    return 0;
}

